(* ========================================================================= *)
(*                                                                           *)
(*                         Flip_gate.ml                                  *)
(*                                                                           *)
(*                                                                           *)
(*                                                                           *)
(* Last update: February 19, 2016                                            *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)


needs "Hadamard_gate.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)
(*--------------------**********************------------------*) 	


let FLip_In_Outputs = define 
`FLip_In_Outputs ((a:sm^N), (c1:sm^N))  <=> 
tensor 1 ((lambda i. vac (a$1)):bqs^N) =
tensor 2 ((lambda i. if i = 1 then fock  (a$1) 1  else vac (a$1)):bqs^N)   	   
/\ tensor 1 ((lambda i. fock (a$1) 1):bqs^N) =
tensor 2 ((lambda i. if i = 2 then fock  (a$2) 1  else vac (a$1)):bqs^N)   
/\ tensor 2 ((lambda j. if j = 1 then fock  (c1$1) 1 else vac (c1$2)):bqs^N)
= tensor 1 ((lambda i. vac (c1$1)):bqs^N) /\ 
tensor 2 ((lambda j. if j = 2 then fock  (c1$2) 1 else vac (c1$2)):bqs^N) 
= tensor 1 ((lambda i. fock (c1$1) 1):bqs^N)  `;; 


let flip_model = 
new_definition ` flip_model ((a:sm^N), (c1:sm^N), ten) <=> 
!(a1:sm^N) (c:sm^N). Hadamard_GATE ((a:sm^N), (c:sm^N), ten) 
/\ c$1 = a1$1 /\ PHASE_SHIFTER (ten,pi,c$2,2,a1$2,2) /\
Hadamard_GATE ((a1:sm^N), (c1:sm^N), ten)  `;;


let Flip_GATE =  
new_definition `Flip_GATE ((a:sm^N), (c1:sm^N), ten)  <=> 
flip_model (a,c1,ten)  /\ FLip_In_Outputs (a, c1) `;;


let FlIP_0 = 
prove(`!(a:sm^N) (c1:sm^N) (ten:qop^N->(real^N->complex)-> (real^N->complex)).
is_tensor ten /\  Flip_GATE (a, c1,ten)   ==>
tensor 2  ((lambda i. if i = 1 then fock  (a$1) 1  else vac (a$1)):bqs^N)=	  
tensor 2  ((lambda i. if i = 2 then fock  (c1$2) 1  else vac (c1$2)):bqs^N)   `, 
REWRITE_TAC[LEFT_IMP_FORALL_THM;LEFT_AND_FORALL_THM;
RIGHT_AND_FORALL_THM;Flip_GATE;flip_model] THEN
REPEAT GEN_TAC THEN
MAP_EVERY EXISTS_TAC [`(a1:sm^N)` ; `(c:sm^N)`]
THEN IMP_REWRITE_TAC[HADAMARD_0;HADAMARD_1] THEN
REWRITE_TAC[PHASE_SHIFTER;pos] 
THEN ASM_SIMP_TAC[LET_RULE_L[GSYM COP_SMUL_THM] FOCK_HERM_VAC]  
THEN SIMP_TAC[cop_pow;COP_MUL_RID;ONE;FACT] THEN
SIMP_TAC[GSYM ONE] THEN SIMP_TAC[ MULT_CLAUSES;SQRT_1;
COMPLEX_INV_1;COP_SMUL_LID] THEN
ONCE_REWRITE_TAC[MESON[I_THM] `(if p then (x:bqs) else y) 
= (if p then x else I y)`] THEN 
ONCE_REWRITE_TAC[MESON[] `(if p then f1 x else f2 y) = 
(if p then f1  else f2 ) (if p then  x else  y)`] THEN
ONCE_REWRITE_TAC[CFUN_LAMBDA_APP] THEN		  
SIMP_TAC[keylem;ARITH;COND_ID] THEN
IMP_REWRITE_TAC[GSYM COP_TENSOR_CFUN] THEN
ASM_SIMP_TAC[GSYM CFUN_LAMBDA_APP;COND_RATOR;I_THM;ARITH;
COP_ADD_THM;CNJ_CX;COP_SMUL_THM;pos;COP_TENSOR_CFUN] THEN
ASM_REWRITE_TAC[GSYM PHASE_SHIFTER;GSYM pos] THEN
ONCE_REWRITE_TAC [(MESON [lemma1;PHASE_SHIFTER;
Hadamard_GATE;is_beam_splitter] `
Hadamard_GATE (a1, c1, ten)  <=> 
( Hadamard_GATE (a1, c1, ten)  /\  
vac (a1$2) = vac (a1$1) /\
 cr (a1$2) (vac (a1$2)) = fock (a1$2) 1)  `)]
THEN IMP_REWRITE_TAC[HADAMARD_0;HADAMARD_1] THEN
SIMP_TAC[CEXP_NEG;COMPLEX_MUL_LNEG;COMPLEX_INV_1;
GSYM COMPLEX_NEG_INV;CEXP_II_PI] THEN
SIMP_TAC [CFUN_SMUL_DIS;CFUN_SMUL_ASSOC;
CFUN_SUB_LDISTRIB;GSYM CX_MUL;] THEN
SIMP_TAC [REAL_MUL_LNEG;GSYM CX_ADD;CFUN_SUB_NEG;
CFUN_ADD_AC;GSYM CFUN_SMUL_LNEG;GSYM CFUN_ADD_RDISTRIB;
GSYM CX_MUL; GSYM CX_NEG;GSYM CX_ADD; GSYM CX_SUB;
GSYM CX_INV;CFUN_ADD_AC;CFUN_ADD_LID;CFUN_SMUL_LZERO;
SQRT_DIV;SQRT_1;GSYM REAL_POW_2;REAL_POS;SQRT_POW_2;
REAL_MUL_SYM;GSYM real_sub;
GSYM REAL_MUL_ASSOC;REAL_MUL_AC;REAL_POW_DIV;
REAL_POW_ONE;REAL_MUL_RNEG;CFUN_SMUL_LZERO;
REAL_MUL_RID;REAL_ADD_LINV;REAL_MUL_LZERO;
REAL_DIV_RMUL;REAL_SUB_RNEG;CFUN_ADD_RID;
REAL_MUL_RZERO;REAL_ADD_LID;REAL_ADD_RID;
REAL_SUB_RZERO;REAL_SUB_REFL;REAL_NEG_NEG;
REAL_FIELD ` &1 / &2 + &1 / &2 = &1`;CFUN_SMUL_LID]);;







let FlIP_1 = 
prove( `!(a:sm^N) (c1:sm^N) (ten:qop^N->(real^N->complex)-> (real^N->complex)).
is_tensor ten /\  Flip_GATE (a, c1, ten) ==>
tensor 2  ((lambda i. if i = 2 then fock  (a$2) 1  else vac (a$1)):bqs^N)=	  
tensor 2  ((lambda i. if i = 1 then fock  (c1$1) 1  else vac (c1$2)):bqs^N)   `, 
REWRITE_TAC[LEFT_IMP_FORALL_THM;LEFT_AND_FORALL_THM;
RIGHT_AND_FORALL_THM;Flip_GATE;flip_model] THEN
REPEAT GEN_TAC THEN
MAP_EVERY EXISTS_TAC [`(a1:sm^N)` ; `(c:sm^N)`] THEN
IMP_REWRITE_TAC[HADAMARD_0;HADAMARD_1] THEN
REWRITE_TAC[PHASE_SHIFTER;pos] 
THEN ASM_SIMP_TAC[LET_RULE_L[GSYM COP_SMUL_THM] FOCK_HERM_VAC]  
THEN SIMP_TAC[cop_pow;COP_MUL_RID;ONE;FACT] THEN
SIMP_TAC[GSYM ONE] THEN SIMP_TAC[ MULT_CLAUSES;
SQRT_1;COMPLEX_INV_1;COP_SMUL_LID] THEN
ONCE_REWRITE_TAC[MESON[I_THM] `(if p then (x:bqs) else y) 
= (if p then x else I y)`] THEN 
ONCE_REWRITE_TAC[MESON[] `(if p then f1 x else f2 y) 
= (if p then f1  else f2 ) (if p then  x else  y)`]
THEN ONCE_REWRITE_TAC[CFUN_LAMBDA_APP] THEN		  
SIMP_TAC[keylem;ARITH;COND_ID] THEN
IMP_REWRITE_TAC[GSYM COP_TENSOR_CFUN] THEN
ASM_SIMP_TAC[GSYM CFUN_LAMBDA_APP;COND_RATOR;I_THM;ARITH;
COP_ADD_THM;CNJ_CX;COP_SMUL_THM;pos;COP_TENSOR_CFUN] THEN
ASM_REWRITE_TAC[GSYM PHASE_SHIFTER;GSYM pos] THEN
ONCE_REWRITE_TAC [(MESON [lemma1;PHASE_SHIFTER;
Hadamard_GATE;is_beam_splitter] `
Hadamard_GATE (a1, c1, ten)  <=> 
( Hadamard_GATE (a1, c1, ten) /\  
vac (a1$2) = vac (a1$1) /\ cr (a1$2) (vac (a1$2)) =
 fock (a1$2) 1) `)] THEN
IMP_REWRITE_TAC[HADAMARD_0;HADAMARD_1]  THEN
SIMP_TAC[CEXP_NEG;COMPLEX_MUL_LNEG;COMPLEX_INV_1;
GSYM COMPLEX_NEG_INV;CEXP_II_PI] THEN
SIMP_TAC [CFUN_SMUL_DIS;CFUN_SMUL_ASSOC;
CFUN_SUB_LDISTRIB;GSYM CX_MUL;CFUN_SUB_AC] THEN
SIMP_TAC [REAL_MUL_LNEG;GSYM CX_ADD;CFUN_SUB_NEG;
CFUN_ADD_AC;GSYM CFUN_SMUL_LNEG;GSYM CFUN_ADD_RDISTRIB;
GSYM CX_MUL; GSYM CX_NEG;GSYM CX_ADD; GSYM CX_SUB;
GSYM CX_INV;CFUN_ADD_AC;CFUN_ADD_LID;CFUN_SMUL_LZERO;
SQRT_DIV;SQRT_1;GSYM REAL_POW_2;REAL_POS;SQRT_POW_2;
REAL_MUL_SYM;GSYM real_sub;
GSYM REAL_MUL_ASSOC;REAL_MUL_AC;REAL_POW_DIV;
REAL_POW_ONE;REAL_MUL_RNEG;CFUN_SMUL_LZERO;
REAL_MUL_RID;REAL_ADD_LINV;REAL_MUL_LZERO;
REAL_DIV_RMUL;REAL_SUB_RNEG;CFUN_ADD_RID;
REAL_MUL_RZERO;REAL_ADD_LID;REAL_ADD_RID;
REAL_SUB_RZERO;REAL_SUB_REFL;REAL_NEG_NEG;
REAL_FIELD ` &1 / &2 + &1 / &2 = &1`;CFUN_SMUL_LID;CFUN_ARITH 
`!(f1:cfun) (b:complex) (a:complex) (c:cfun). 
 a % f1 + b % f1  + c = (a+b)% f1 + c`]);;
