(* ========================================================================= *)
(*                                                                           *)
(*                          Hadamard_gate.ml                              *)
(*                                                                           *)
(*                                                                           *)
(*                                                                           *)
(* Last update: February 19, 2016                                            *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)


needs "CZ_gate.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)
(*--------------------**********************------------------*) 	  
let CEXP_II_PI = prove(`cexp (ii * Cx(pi)) = --Cx(&1)`,
    REWRITE_TAC[CEXP_EULER;GSYM CX_COS;GSYM CX_SIN;COS_PI;
    SIN_PI;COMPLEX_ADD_RID;COMPLEX_MUL_RZERO;CX_NEG]);;  
	  	   


let Hadamard_In_Output = define 
`Hadamard_In_Output ((a:sm^N), (c:sm^N))  <=> 
tensor 1 ((lambda i. vac (a$1)):bqs^N) =
tensor 2 ((lambda i. if i = 1 then fock  (a$1) 1  else vac (a$1)):bqs^N)   	   
/\ tensor 1 ((lambda i. fock (a$1) 1):bqs^N) =
tensor 2 ((lambda i. if i = 2 then fock  (a$2) 1  else vac (a$1)):bqs^N)  	   
/\ tensor 2 ((lambda j. if j = 1 then fock  (c$1) 1 else vac (c$2)):bqs^N)
= tensor 1 ((lambda i. vac (c$1)):bqs^N) /\ 
tensor 2 ((lambda j. if j = 2 then fock  (c$2) 1 else vac (c$2)):bqs^N) 
= tensor 1 ((lambda i. fock (c$1) 1):bqs^N)  `;; 


let Hadamard_GATE = 
  new_definition `Hadamard_GATE ((a:sm^N), (c:sm^N), ten)  <=> 
!(b:sm^N).  is_beam_splitter (Cx(sqrt(&1/ &2)),--Cx(sqrt(&1/ &2)),
Cx(sqrt(&1/ &2)),Cx(sqrt(&1/ &2)),ten,a$1,1,a$2,2,c$1,1,b$2,2)
/\ PHASE_SHIFTER (ten,pi,b$2,2,c$2,2) /\            
 Hadamard_In_Output (a, c)`;;
	   

let HADAMARD_1 = 
 prove( `!a c (ten:qop^N->(real^N->complex)-> (real^N->complex)) .
 is_tensor ten /\ Hadamard_GATE ((a:sm^N), (c:sm^N), ten) ==>
tensor 2 ((lambda i. if i = 2 then fock  (a$2) 1  else vac (a$1)):bqs^N)=	  
 Cx (sqrt (&1 / &2)) %  
tensor 2 (lambda j. if j = 1 then fock  (c$1) 1 else vac (c$2)) -  
Cx (sqrt (&1 / &2)) % 
 tensor 2 (lambda j. if j = 2 then fock  (c$2) 1 else vac (c$2)) `,
REWRITE_TAC[LEFT_IMP_FORALL_THM;LEFT_AND_FORALL_THM;
RIGHT_AND_FORALL_THM;Hadamard_GATE]
THEN REPEAT GEN_TAC THEN MAP_EVERY EXISTS_TAC [`(b:sm^N)`] THEN	  	 
REWRITE_TAC[is_beam_splitter;PHASE_SHIFTER;pos] THEN REPEAT STRIP_TAC
THEN ASM_SIMP_TAC[LET_RULE_L[GSYM COP_SMUL_THM] FOCK_HERM_VAC]  THEN 
MULTI_MODE_DECOMPOSE_TAC THEN SIMP_TAC[cop_pow;COP_MUL_RID;ONE;FACT] THEN
SIMP_TAC[GSYM ONE] THEN SIMP_TAC[ MULT_CLAUSES;SQRT_1;COMPLEX_INV_1;
COP_SMUL_LID] THEN ASM_SIMP_TAC[CNJ_CX;GSYM CX_NEG;COP_ADD_LDISTRIB] 
THEN SIMP_TAC[CEXP_NEG;COMPLEX_MUL_LNEG;COMPLEX_INV_1;GSYM COMPLEX_NEG_INV;
COP_SUB_THM;COP_SMUL_THM;CEXP_II_PI;COMPLEX_MUL_RNEG;COMPLEX_MUL_RID;
COP_SMUL_ASSOC;GSYM COP_SUB;COP_SMUL_LNEG]);;


let HADAMARD_0 = 
 prove( `!a c (ten:qop^N->(real^N->complex)-> (real^N->complex)).
is_tensor ten /\ Hadamard_GATE ((a:sm^N), (c:sm^N), ten) ==>
tensor 2  ((lambda i. if i = 1 then fock  (a$1) 1  else vac (a$1)):bqs^N)=	  
Cx (sqrt (&1 / &2)) %  
tensor 2 (lambda j. if j = 2 then fock  (c$2) 1 else vac (c$2)) +  
Cx (sqrt (&1 / &2)) % 
tensor 2 (lambda j. if j = 1 then fock  (c$1) 1 else vac (c$2)) `,
REWRITE_TAC[LEFT_IMP_FORALL_THM;LEFT_AND_FORALL_THM;
RIGHT_AND_FORALL_THM;Hadamard_GATE]
THEN REPEAT GEN_TAC THEN MAP_EVERY EXISTS_TAC [`(b:sm^N)`] THEN	  	 
REWRITE_TAC[is_beam_splitter;PHASE_SHIFTER;pos] THEN REPEAT STRIP_TAC
THEN ASM_SIMP_TAC[LET_RULE_L[GSYM COP_SMUL_THM] FOCK_HERM_VAC]  
THEN MULTI_MODE_DECOMPOSE_TAC
THEN SIMP_TAC[cop_pow;COP_MUL_RID;ONE;FACT] THEN
SIMP_TAC[GSYM ONE] THEN SIMP_TAC[ MULT_CLAUSES;SQRT_1;COMPLEX_INV_1;
COP_SMUL_LID] THEN
ASM_SIMP_TAC[CNJ_CX;GSYM CX_NEG;COP_ADD_LDISTRIB] THEN
SIMP_TAC[CEXP_NEG;COMPLEX_MUL_LNEG;COMPLEX_INV_1;GSYM COMPLEX_NEG_INV;
COP_SUB_THM;COP_SMUL_THM;CEXP_II_PI;COMPLEX_MUL_RNEG;COMPLEX_MUL_RID;
COP_SMUL_ASSOC;GSYM COP_SUB;COP_SMUL_LNEG] THEN SIMP_TAC[COP_SMUL_THM;
COP_ADD_THM;GSYM COP_SMUL_LNEG] THEN CFUN_ARITH_TAC);;